/*
	SNES region patcher implementation
	Mic, 2010
*/

#include "SnesRegionPatcher.h"

// Check for this
const unsigned char NTSC_PATTERN0[] = {0xAD, 0x3F, 0x21, 0x29, 0x10};				// lda $213f / and #$10
// Replace with this
const unsigned char NTSC_TO_PAL_PATTERN0[] = {0xAD, 0x3F, 0x21, 0xA9, 0x00};		// lda $213f / lda #0

const unsigned char NTSC_PATTERN1[] = {0xAF, 0x3F, 0x21, 0x00, 0x89, 0x10};			// lda.l $00213f / bit #$10
const unsigned char NTSC_TO_PAL_PATTERN1[] = {0xAF, 0x3F, 0x21, 0x00, 0xA9, 0x00};	// lda.l $00213f / lda #0

const unsigned char* const NTSC_PATTERNS[] =
{
	NTSC_PATTERN0,
	NTSC_PATTERN1,
	NULL
};

const int NTSC_PATTERN_LENGTHS[] = {5, 6};

// Not implemented yet
const unsigned char* const PAL_PATTERNS[] =
{
	NULL
};

const int PAL_PATTERN_LENGTHS[] = {0};


const unsigned char* const *SEARCH_PATTERNS[2] =
{
	NTSC_PATTERNS,
	PAL_PATTERNS
};

const int* const PATTERN_LENGTHS[2] =
{
	NTSC_PATTERN_LENGTHS,
	PAL_PATTERN_LENGTHS
};

const unsigned char* const NTSC_TO_PAL_PATTERNS[] =
{
	NTSC_TO_PAL_PATTERN0,
	NTSC_TO_PAL_PATTERN1,
	NULL
};

// Not implemented yet
const unsigned char* const PAL_TO_NTSC_PATTERNS[] =
{
	NULL
};

const unsigned char* const *REPLACEMENT_PATTERNS[2] =
{
	PAL_TO_NTSC_PATTERNS,
	NTSC_TO_PAL_PATTERNS
};


int SnesRegionPatcher::PatternReplace(const unsigned char *const *searchPatterns,
									  const unsigned char *const *replacementPatterns,
									  const int *patternLengths,
									  char *dataToPatch,
									  int dataSize)
{
	int numPatches = 0;
	int *stillValid;
	int numPatterns;
	int dataPos;

	if ((searchPatterns != NULL) && (replacementPatterns != NULL))
	{
		for (numPatterns = 0; searchPatterns[numPatterns] != NULL; numPatterns++);
		stillValid = new int[numPatterns + 1];

		dataPos = 0;

		while (dataPos < dataSize)
		{
			stillValid[numPatterns] = numPatterns;
			for (int i = 0; i < numPatterns; i++)
			{
				stillValid[i] = 0;
			}
			while (stillValid[numPatterns] > 0)
			{
				for (int i = 0; i < numPatterns; i++)
				{
					if ((stillValid[i] >= 0) && (stillValid[numPatterns] > 0))
					{
						if (dataPos + stillValid[i] < dataSize)
						{
							if ((unsigned char)dataToPatch[dataPos + stillValid[i]] == (unsigned char)searchPatterns[i][stillValid[i]])
							{
								if ((++stillValid[i]) >= patternLengths[i])
								{
									// We've found a match
									for (int r = 0; r < patternLengths[i]; r++)
									{
										dataToPatch[dataPos + r] = replacementPatterns[i][r];
									}
									dataPos += patternLengths[i] - 1;
									stillValid[numPatterns] = 0;
									numPatches++;
									break;
								}
							}
							else
							{
								stillValid[i] = -1;
								stillValid[numPatterns]--;
							}
						}
						else
						{
							stillValid[i] = -1;
							stillValid[numPatterns]--;
						}
					}
				}
			}
			dataPos++;
		}

		delete [] stillValid;
	}

	return numPatches;
}


int SnesRegionPatcher::Patch(const int newRegion, const char *smcFilename, const char *outputFilename)
{
	std::ifstream smcFile;
	std::ifstream::pos_type fileSize;
	char *fileBuffer;
	int numPatches;
	int oldRegion = 1 - newRegion;

	if (outputFilename == NULL)
	{
		outputFilename = smcFilename;
	}

	smcFile.open(smcFilename, std::ios::in | std::ios::binary | std::ios::ate);
    if (smcFile.is_open())
    {
		fileSize = smcFile.tellg();
		fileBuffer = new char[fileSize];
		smcFile.seekg (0, std::ios::beg);
		smcFile.read(fileBuffer, fileSize);
		smcFile.close();
		numPatches = PatternReplace(SEARCH_PATTERNS[oldRegion],
			                        REPLACEMENT_PATTERNS[newRegion],
									PATTERN_LENGTHS[oldRegion],
									fileBuffer,
									fileSize);
		if ((numPatches > 0) || (strcmp(smcFilename, outputFilename) != 0))
		{
			std::ofstream outputFile(outputFilename, std::ios::out | std::ios::binary);
			outputFile.write(fileBuffer, fileSize);
			outputFile.close();
		}
		delete [] fileBuffer;
	}
	else
	{
		numPatches = -1;
	}

	return numPatches;
}
